/* Recursive drawing of Koch curves.  */

#include <fork.h>
#include <assert.h>
#include <io.h>
#include <stdlib.h>
#include <math.h>
#include <graphic.h>

#define PIXELS 252      /*extents of generated X-Bitmap picture*/

sh double factor = 0.5;
sh int MAXLEVEL = 5;

/* output routines: */

async void seq_koch ( pr int startx, pr int starty, 
                      pr int stopx, pr int stopy, pr int level )
{
 if (level >= MAXLEVEL) {   /*reach the limit of recursion*/
    seq_line( startx, starty, stopx, stopy, 0xc0c0c0, 1 );
    return;
 } /* else: another recursion step */
 else {
   int x[5], y[5], i;
   float dx, dy;
   seq_line( startx, starty, stopx, stopy, 0xc0c000+(level+2)*0x10, 1 );
   dx = (float)(stopx - startx);
   dy = (float)(stopy - starty);
   /* linear interpolation: */
   x[0] = startx;
   y[0] = starty;
   x[1] = startx + (0.33333*dx); 
   y[1] = starty + (0.33333*dy);
   x[2] = startx + dx/2 - (int)(factor * (float)dy);
   y[2] = starty + dy/2 + (int)(factor * (float)dx);
   x[3] = startx + (0.66666*dx); 
   y[3] = starty + (0.66666*dy);
   x[4] = stopx;
   y[4] = stopy;
   for (i=0; i<4; i++)
     seq_koch( x[i], y[i], x[i+1], y[i+1], level + 1 );
 }
}


sync void koch ( sh int startx, sh int starty, 
                 sh int stopx, sh int stopy, sh int level )
{
 if (level >= MAXLEVEL) {   /*reach the limit of recursion*/
    line( startx, starty, stopx, stopy, 0xc0c0c0 , 1 );
 }
 else { /* another recursion step */
   sh int x[5], y[5];
   sh float dx, dy;
   pr int j;
   line( startx, starty, stopx, stopy, (level+1)*0x10, 1 );
   dx = stopx - startx;
   dy = stopy - starty;
   /* linear interpolation: */
   x[0] = startx;
   y[0] = starty;
   x[1] = startx + (int)(dx*0.33333); 
   y[1] = starty + (int)(dy*0.33333);
   x[2] = startx + dx/2 - (int)(factor * dy);
   y[2] = starty + dy/2 + (int)(factor * dx);
   x[3] = startx + (dx*0.66666); 
   y[3] = starty + (dy*0.66666);
   x[4] = stopx;
   y[4] = stopy;
   if (#<4)
     farm
       forall( j, 0, 4, # )
         seq_koch( x[j], y[j], x[j+1], y[j+1], level + 1 );
   else
     fork ( 4; @=$$%4; )
       koch( x[@], y[@], x[@+1], y[@+1], level + 1 );
 }
}
 


void main( void )
{
 if (__PROC_NR__==0) printf("Clearing picture ...\n");
 start {
    init_pict( PIXELS+1, PIXELS+1 );
    clear_pixels( 0 );
    seq {
       printf("Enter MAXLEVEL = ");
       scanf("%d", &MAXLEVEL );
       printf("Enter factor = ");
       scanf("%f", &factor );
       printf("\nGenerating Koch curve (%f) of level %d...\n", factor, MAXLEVEL);
    }
    initTracing(100000);
    startTracing();
    koch ( 1, PIXELS/4, PIXELS-1, PIXELS/4, 1 );
    stopTracing();
    writeTraceFile("koch.trv", "Drawing Koch curves");
    write_pixmap( "KOCH" );
 }
 exit(0);
}
