/* N-QUEENS problem, Fork95 (V1.7) implementation 9710 C. Kessler */

#include <fork.h>
#include <syscall.h>
#include <assert.h>
#include <io.h>

sh int N;            /* the number of queens */

sh simple_lock screen = 0;  /*lock for screen output*/

sh int solutions = 0;   /*counter for number of solutions*/

/** q is allocated with N entries, n<N.
 *  It has already been tested whether column index j occurs in q[].
 */
void seq_test_queen( int *q, int n, int j )
{
 int i, l;
 // two queens are on same diagonal iff col1-row1==col2-row2
 // two queens are on same counterdiagonal iff col1+row1==col2+row2
 for (i=0; i<n; i++) {
   if (j-q[i] == n-i) return;   /*fail*/
   if (q[i]-j == n-i) return;   /*fail*/
 }
 // now position j is feasible.
 q[n]=j;

 if (n<N-1) { // not yet arrived at last row:
    for (i=0; i<N; i++) {   // try all column indices i
      int thr, cq;
      for (cq=0,thr=0; cq<n; cq++)   // loop over q[]
        if (q[cq]==i) {thr=1;break;} // column i threatened by queen cq
      if (thr || (i==j)) continue;   // column i threatened by a queen
      seq_test_queen( q, n+1, i );   // recursive call for row n+1
    }
 }
 else  // final row
    syncadd( &solutions, 1 );
}

/** Queens in row 0...n-1 have been positioned, 
 *  their column indices are stored in q[0:n-1].
 *  A new queen may be set at position (n,j).
 *  Test queen at (n,j) for diagonal threatening,
 *  and do recursive tests for rows n+1...N-1:
 */
sync void test_queen( sh int *q, sh int n, sh int j )
{
 sh int p = 0;                /* number of available procs */
 sh int *qq;
 sh int *cand;
 sh int cands = 0;
 sh int feasible = 1;         /*flag to be cleared if j is unfeasible*/
 int i, l;

 $ = mpadd( &p, 1 );   /*renumber $ and determine p*/
 if (p==1) {
    qq = (int *)shalloc( N * sizeof(int) );
    farm {
      for (i=0; i<n; i++) qq[i] = q[i];
      seq_test_queen( qq, n, j );
    }
    shallfree();
    return;
 }
 // it has already been tested whether column index j occurs in q[] */
 farm
  forall (i,0,n,p) {
   // two queens are on the same diagonal 
   // iff either row2-row1 == col1-col2 or row2-row1==col2-col1:
   if (j-q[i] == n-i)  feasible = 0;   // -> fail
   if (q[i]-j == n-i)  feasible = 0;   // -> fail
  }
 if (!feasible) return;

 if (n<N-1) {  // not yet arrived at last row.
    // generate list of N-n candidates for row n+1:
    cand = (int *)shalloc( (N-n) * sizeof(int) );
    farm
      forall (i,0,N,p) { /* in parallel try all column indices i */
        int pos, cq, thr;
        for (cq=0,thr=0; cq<n; cq++) /* loop over q[] */
          if (q[cq]==i) {thr=1;break;}  /* column i threatened by queen cq */
        if (!thr && (i!=j)) {         /*column i not threatened by a queen */
          pos = mpadd( &cands, 1 );
          cand[pos] = i;   /*found feasible candidate*/
        }
      }
    if (!cands) return;
    if (p < cands) { // p is too small. Apply strip mining
      sh int iter = 0;
      int *myqq;
      int k;
      farm {   // one temporary array per processor:
        myqq = (int *)shmalloc( N * sizeof(int) );
        for (l=0; l<n; l++)  myqq[l] = q[l];
        myqq[n]=j;
        FORALL( k, &iter, 0, cands, 1 )
          seq_test_queen( myqq, n+1, cand[k] );
        shfree( myqq );
      }
    }
    else { // p is large enough. Make one subgroup for each candidate
      qq = (int *)shalloc( (n+1) * sizeof(int) );
      farm forall(l,0,n,p)  qq[l] = q[l];
      qq[n]=j;
      fork( cands; @=$%cands; )
        test_queen( qq, n+1, cand[@] );
      shallfree();
    }
 }
 else  // arrived at last row: print the solution
    syncadd( &solutions, 1 );
}

void main( void )
{
 if (__PROC_NR__==0) {
    N = __STARTED_PROCS__;
    if (N > __STARTED_PROCS__) {
      exit(1);
    }
 }


 start {

   fork( N; @=$%N; ) {
      test_queen( NULL, 0, @ );
   }
 }
 exit(0);
}
