/* SBPRAM-FORK header fork.h     CWK 940628 */
/* last change: Dec 4, 1995 */

#ifndef __FORK_HEADER__
#define __FORK_HEADER__
/* required declarations for this header file: */
extern straight void *reltoabs( pr void * );
extern straight void exit( int );  // synchrony does not matter any more

extern pr int __mp_dummyvar;        /* make multiprefix work correctly */
//#define mpadd(a,b) (__mp_dummyvar = mpadd((a),b))
//#define mpmax(a,b) (__mp_dummyvar = mpmax((a),b))
//#define mpand(a,b) (__mp_dummyvar = mpand((a),b))
//#define mpor(a,b) (__mp_dummyvar = mpor((a),b))
                                    /* bug fix to guarantee side effect */
extern straight void syncadd (int *, int);
extern straight void syncadd_m0 (int *, int);
extern straight void syncadd_m1 (int *, int);
extern straight void syncmax (int *, int);
extern straight void syncmax_m0 (int *, int);
extern straight void syncmax_m1 (int *, int);
extern straight void syncand (int *, int);
extern straight void syncand_m0 (int *, int);
extern straight void syncand_m1 (int *, int);
extern straight void syncor (int *, int);
extern straight void syncor_m0 (int *, int);
extern straight void syncor_m1 (int *, int);

extern sh const int __STARTED_PROCS__; /* number of started processors */
                                       /* generally unknown at compile time */
extern pr const int __PROC_NR__;       /* original consecutive process ID,
                                        * starting with 0 */
#define NULL 0

extern void *malloc( unsigned int );   /*allocate private heap space. No free() */
extern sync void *shalloc( sh unsigned int );
            /*allocate shared heap space. No explicit free(). Use shallfree()*/
extern sync void shallfree( void ); 
              /*release all blocks shalloc()ed in the current (sync) function*/ 

extern async void *shmalloc( unsigned int );
extern async void *alloc( unsigned int );
/* allocate permanent shared heap space:
   alloc for new space, shmalloc for reuse */

extern async void shfree( void * );     /*dummy permanent shared heap free()*/
extern straight int pravail( void );     /* returns #free priv. memory words */
extern straight int shavail( void );     /* returns #free shar. memory words */

extern straight unsigned int getct( void );   /* get global time counter, 1cc=4msec */
extern straight unsigned int getnr( void );   /* get hardware proc ID */

extern straight int groupsize( void );
extern straight int parentgroupsize( void );
extern async int _groupsize( unsigned int ); /*new 970901*/

extern int *__TICKETP__;   /* points to my current busTicket counter 990128*/
extern int *__RANKP__;   /* points to my current busRank counter 990128*/
#define __NUM_PR__ (*__TICKETP__)
#define __RANK__ (*__RANKP__)


/* =================  L O C K S :  ======================================== */

/* Lock types, as proposed by J. Roehrig in his Master Thesis, 1995,
 *             Saarbruecken University, taken from his p4 library implementation
 *             for the SB-PRAM: 
 *
 * 4 different locks are provided:
 *
 *     simple locks, safe locks, fair locks & reader/writer-locks
 *
 * simple locks work faster than safe locks and fair locks
 * safe locks provide some features useful for debugging
 * fair locks guarantee access to a lock in demand order (when using simple
 *   or safe locks, processors with low IDs will always get access before
 *   processors with high IDs because of the multiprefix order)
 *
 * reader/writer-locks provide the following mechanism:
 * - several readers can own the lock simultaneously
 * - a reader and a writer can't own the lock simultaneously
 * - two different writers can't own the lock simultaneously
 *
 * the locking mechanism is based on the following functions:
 * (MODULO is the modulo bit which toggels on each round)
 *
 * initialization of a lock  - must be called before any attempt to catch
 * a lock and should only be executed by one process (master)
 *
 *    void simple_lock_init(simple_lock *lock)
 *    void safe_lock_init(safe_lock *lock)
 *    void fair_lock_init(fair_lock *lock)
 *    void rw_lock_init(rw_lock *lock)
 *
 * catch a lock - lock must be initialized
 * (will loop until the lock could be locked - there's no timeout !) 
 *
 *    void simple_lockup(simple_lock *lock)
 *    void safe_lockup(safe_lock *lock)
 *    void fair_lockup(fair_lock *lock)
 *    void rw_lockup(rw_lock *lock, PTYPE)
 *         where "PTYPE" in {SBP_RW_READER, SBP_RW_WRITER)
 *
 * unlock a lock - executing process must be owner of the lock
 *
 *  simple version/fair version
 *  ---------------------------  
 *  (theoretically it's possible that a proccess that doesn't hold a lock l
 *  can unlock l - this will lead to unpredictable results and must be
 *  avoided by the programmer)
 * 
 *  this function will always return 0 (for compatibility with the safe version)
 *
 *    int simple_unlock(lock *lock)
 *    int fair_unlock(fair_lock *lock)
 * 
 *  safe version
 *  ------------ 
 *  (return code:
 *      0 - unlock was successful
 *      1 - tried to unlock a lock that wasn't locked
 *      2 - tried to unlock a lock whose owner was another process
 *   in case 1 and 2, the state of the lock isn't changed)
 *
 *    int safe_unlock(safe_lock *lock)
 * 
 *  reader/writer-lock
 *  ------------------
 * 
 *   void rw_unlock(rw_lock *lock, PTYPE, int wait)
 *        "PTYPE" in {SBP_RW_READER, SBP_RW_WRITER)
 *        "wait" is the delay a writer waits before freeing the rw-lock
 *               for other writers after having freed it for readers
 *               (delay corresponds to wait*3 SBPRAM-rounds
 */

typedef int simple_lock;     /* simple locks */

typedef struct {             /* fair locks */
   int nextnum;
   int actnum;
} fair_lock;

typedef struct {             /* safe locks */
   int cell;                     /* memory cell holding the lock */
   int owner;                    /* process ID of lock owner */
} safe_lock;

typedef struct {             /* reader-writer-locks */
   unsigned int readercounter;   /* Bits 0-29 used as reader counter */
                                 /* Bit 30 is used as writer flag */
   fair_lock writerlock;         /* current and next number */
} rw_lock;

typedef struct {             /* reader-writer-locks */
   unsigned int readercounter;   /* Bits 0-29 used as reader counter */
                                 /* Bit 30 is used as writer flag */
   fair_lock writerlock;         /* current and next number */
   int deleteflag;               /* if 1, this lock is going to be deleted */
} rwd_lock;

#define RW_READ 0
#define RW_WRITE 1
#define RW_DELETE 2

/* the following routines are implemented in forklib2.asm */

typedef simple_lock *SimpleLock;
extern simple_lock *new_SimpleLock( void );
#define simple_lock_init(pslock) *(pslock)=0
extern void simple_lockup( simple_lock * );
#define simple_unlock(pslock) *(pslock)=0

typedef fair_lock *FairLock;
extern fair_lock *new_FairLock( void );
extern void fair_lock_init(fair_lock *lock);
extern void fair_lockup( fair_lock * );
extern int fair_lockup_delflag( fair_lock *, int * );
extern void fair_unlock( fair_lock * );

extern safe_lock *new_safe_lock( void );
#define safe_lock_init(slock) slock={0,-1}
extern void safe_lockup( safe_lock * );
extern int safe_unlock( safe_lock * );   /* returns an error code */

/* the following routines are implemented in async.c: */
typedef rw_lock *RWLock;
extern rw_lock *new_RWLock( void );
extern void rw_lock_init( rw_lock * );
extern void rw_lockup( rw_lock *, int );
extern void rw_unlock( rw_lock *, int, int );

typedef rwd_lock *RWDLock;
extern rwd_lock *new_RWDLock( void );
extern void rwd_lock_init( rwd_lock * );
extern int rwd_lockup( rwd_lock *, int );  /* returns 0 if failed, 1 if lock acquired */
extern void rwd_unlock( rwd_lock *, int, int );


/* critical sections: some useful macros */

#define _enter_crit_sect(adr) simple_lockup(adr) 
#define _exit_crit_sect(adr) simple_unlock(adr)

#if 0
/*other alternatives:*/
#define _enter_crit_sect(adr) fair_lockup(adr)
#define _exit_crit_sect(adr) fair_unlock(adr)
/*or*/
#define _enter_crit_sect(adr) safe_lockup(adr)
#define _exit_crit_sect(adr) safe_unlock(adr)
#endif

/*#define DEBUG_CRIT_SECT      /* print locking and unlocking events on screen */
#ifdef DEBUG_CRIT_SECT
#define enter_crit_sect(adr) {\
      extern int pprintf ( char *, ... );\
      pprintf("enter_crit_sect, line %d", __LINE__);\
      _enter_crit_sect(adr);}
#define exit_crit_sect(adr) {\
      extern int pprintf ( const char *, ... );\
      pprintf("exit_crit_sect, line %d", __LINE__);\
      _exit_crit_sect(adr);}
#else
#define enter_crit_sect(adr) _enter_crit_sect(adr);
#define exit_crit_sect(adr) _exit_crit_sect(adr);
#endif


// ======= Macros for parallel loops: ============================

// one-dimensional, statically scheduled parallel loop:

#define Forall(i,lb,ub,st,p) for(i=$$+(lb);i<(ub);i+=(st)*(p))

// a variant of Forall for the frequent case of unit-stride loops:

#define forall(i,lb,ub,p) for(i=$$+(lb);i<(ub);i+=(p))

// two-dimensional, flattened, statically scheduled parallel loop:

#define Forall2(i,j,ij,l1,u1,s1,l2,u2,s2,p) \
    for(ij=$$,i=(ij/(((u2)-(l2)+(s2)-1)/(s2)))*(s1)+(l1), \
     j=(ij%(((u2)-(l2)+(s2)-1)/(s2)))*(s2)+(l2); \
    ij<(((u1)-(l1)+(s1)-1)/(s1)) * (((u2)-(l2)+(s2)-1)/(s2)); \
    ij+=(p),i=(ij/(((u2)-(l2)+(s2)-1)/(s2)))*(s1)+(l1), \
     j=(ij%(((u2)-(l2)+(s2)-1)/(s2)))*(s2)+(l2))

// a variant of forall2 for the frequent case of unit strides
// in both dimensions (saves 14 cc per iteration, compared to forall2):

#define forall2(i,j,ij,l1,u1,l2,u2,p) \
    for(ij=$$,i=ij/((u2)-(l2))+(l1),j=ij%((u2)-(l2))+(l2); \
    ij<((u1)-(l1))*((u2)-(l2)); \
    ij+=(p),i=ij/((u2)-(l2))+(l1),j=ij%((u2)-(l2))+(l2))

// one-dimensional, dynamically scheduled parallel loop:

#define FORALL( IVAR, PSHIVAR, LB, UB, ST ) \
  for( *(PSHIVAR)=(LB), barrier, IVAR=mpadd((PSHIVAR),(ST)); \
       IVAR < (UB);  IVAR = mpadd((PSHIVAR),(ST)) )
 



/* ======      B A R R I E R     (group-relative resynchronization) ===== */

straight void barrier( void );   // assembler routine
#define barrier barrier()



/* ======  TRACING ROUTINES:  (see async.c)  990722 */

#define TRACE_END 0
#define TRACE_SPLIT 1
#define TRACE_GROUP 2
#define TRACE_BARRIER_ENTRY 3
#define TRACE_BARRIER_EXIT 4
#define TRACE_LOCKUP_WAIT 5
#define TRACE_LOCKUP_DONE 6
#define TRACE_ENTER_SEND 8
#define TRACE_EXIT_SEND 9
#define TRACE_ENTER_RECV 10 
#define TRACE_EXIT_RECV 11 

extern sync void initTracing( sh int lengthInWords );
extern void *traceEntry( int evnttype, int info ); // user-defined events
   // writes gps if info==-1; returns pointer to info cell
extern sync void startTracing ( void );
extern sync void stopTracing ( void );
extern sync void writeTraceFile( sh char *filename, sh char *title );

extern sh int *_tracebuffer;         // in forklib2.asm
extern sh int _tracebufferlength;
extern sh int _tracestarttime;
extern sh int _tracestoptime;
extern sh int _traceshsize;
extern sh int _tracefirstgps;
extern sh int *_postracebuf;
extern sh int _tracingenabled;

// to avoid spurious errors due to missing linkage information:
#include <io.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#endif
