/* fcc.c  Source fuer fcc - Driver  (c)29.11.1994 C.W. Kessler */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <alloca.h>
#include <argz.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <limits.h>
#include "fcc.h"
#include "version.h"

typedef struct argz {
    char   *arg;
    size_t  len;
} argz_t;
#define INIT_ARGZ { .arg = NULL, .len = 0 }
#define argz_init(__argzp) \
        ({ (__argzp)->arg = NULL, (__argzp)->len = 0; })
#define argz(__argzp) (__argzp)->arg, (__argzp)->len
#define argzp(__argzp) &(__argzp)->arg, &(__argzp)->len

static inline void
argz_clear(argz_t *az)
{
    if (az->arg)
        free(az->arg);
   memset(az, 0, sizeof(*az));
}

static inline void
argz_print(const argz_t *az)
{
    char *c;
    for (c = az->arg; c; c = argz_next(argz(az), c))
        printf("%s%s", c, argz_next(argz(az), c) ? " " : "\n");
}

static inline int
argz_system(const argz_t *az, int vflag)
{
    char **av = NULL;
    size_t s = argz_count(argz(az)) + 1;
    pid_t pid;
    int status;

    if (vflag)
        argz_print(az);

    if (vflag >= 2)
        return 0;

    av = alloca(s * sizeof(char *));
    if (!av)
        return -1;

    argz_extract(argz(az), av);

    switch ((pid = fork())) {
    case 0:
        execvp(av[0], av);
        fprintf(stderr, "can't exec %s\n", av[0]);
        _exit(127);
    case -1:
        fprintf(stderr, "can't execute %s", az->arg);
        return -1;
    default:
        waitpid(pid, &status, 0);
        return WIFEXITED(status) ? 0 : WEXITSTATUS(status);
    }
}

static inline void
argz_add_with_prefix(argz_t *az, const char *prefix, const char *strs[])
{
    char filebuf[PATH_MAX];
    for (; *strs; strs++) {
        snprintf(filebuf, sizeof(filebuf), "%s%s", prefix, *strs);
        argz_add(argzp(az), filebuf);
    }
}

static argz_t compaz;
static argz_t linkaz;
static argz_t prepraz;
static argz_t assaz;
static argz_t tolinkaz;
static argz_t srcaz;
static argz_t adaptaz;
static int Eflag = 0;
static int Sflag = 0;
static int Tflag = 0;
static int cflag = 0;
static int vflag = 0;

/* setze src, compaz, linkopts */
static void
doargs(int argc, char *argv[])
{
    char *s;
    int i;
    for (i = 1; i < argc; i++) {
        s = argv[i];
        if (s[0] == '-') {
            /* Option */
            /* unterscheide Compileroptionen von anderen */
            switch (s[1]) {
            case 'C':
            case 'D':
            case 'I':
            case 'U': /*Pr"aprozessoroptionen:*/
                argz_add(argzp(&prepraz), s);
                break; 
            case 'E':
                Eflag = 1;
                break;
            case 'S':
                Sflag = 1;   /*behalte .s-Files*/
                break;
            case 'T':
                Tflag = 1; /*990504: generate profiling code*/
                argz_add(argzp(&compaz), s);
                break;
            case 'o': /* -o outfilename: */
                argz_add(argzp(&linkaz), s);
                s = argv[++i];
                /* fall through: */
            case 'l': /* -lm Linkoptionen */
                argz_add(argzp(&linkaz), s);
                break;
            case 'c':
                cflag = 1;
                break;
            case 'v':
                vflag++;
                argz_add(argzp(&compaz), s);
                break;
            case 'W': /* arguments for prepr/ass/linker in comma separated list */
                if (s[2] && (s[2] == 'p' || s[2] == 'a' || s[2] == 'l')
                    && s[3] && s[3] == ','
                ) {
                    argz_t *optlist;
                    argz_t cslist = INIT_ARGZ;
                    switch (s[2]) {
                    case 'p':  /* prepr */
                        optlist = &prepraz;
                        break;
                    case 'a':  /* assambler */
                        optlist = &assaz;
                        break;
                    case 'l':  /* linker */
                        optlist = &linkaz;
                        break;
                    }
                    argz_create_sep(&s[4], ',', argzp(&cslist));
                    argz_append(argzp(optlist), argz(&cslist));
                    argz_clear(&cslist);
                    break;
                }
                /* waring for compiler */
                /* fall through */
            default:
                argz_add(argzp(&compaz), s);
                break;
            }
        }
#if 0
       else                /* kein - */
       if (s[0]=='>' || s[0]=='2'&& s[1]=='>') {  /*Umlenkung auf File*/
          Umlenkung = s;
          if (s[1]=='>')   /* trifft 2> und >> */
             if (isalpha(s[2])||ispunct(s[2])||s[2]=='/')  /*Pfad folgt direkt*/
                ;  /*Umlenkung bleibt so*/
             else 
                Umlenkung = append(s, argv[++i]);
          else
             if (isalpha(s[1])||ispunct(s[1])||s[1]=='/')  /*Pfad folgt direkt*/
                ;  /*Umlenkung bleibt so*/
             else
                Umlenkung = append(s, argv[++i]);
       }
#endif
        else {
            /* Sourcefile */
            int n = strlen(s);
            if (s[n - 2] == '.' && s[n - 1] == 'c') {
                argz_add(argzp(&srcaz), s);
            } else if (s[n - 2] == '.' && s[n - 1] == 'o')
                argz_add(argzp(&tolinkaz), s);
            else
                printf("Useless parameter: %s\n", s);
        }
    }
}

static void
add_linkfiles(argz_t *linkaz, argz_t *srcaz)
{
    char *src;
    int len;
    for (src = srcaz->arg; src; src = argz_next(argz(srcaz), src)) {
        len = strlen(src);
        src[len - 1] = 'o';
        argz_append(argzp(linkaz), src, len + 1);
    }
}

void clearerrors(char *srcname)
{
    char errfile[PATH_MAX];
    snprintf(errfile, PATH_MAX, "%.*s.err", (int)(strlen(srcname) - 2), srcname);
    unlink(errfile);
}

int geterrors(char *srcname, int vflag)
{
    FILE *fp;
    int nrerrors = 0;
    char errfile[PATH_MAX];

    snprintf(errfile, PATH_MAX, "%.*s.err", (int)(strlen(srcname) - 2), srcname);
    //printf("scanning %s\n", errfile );
    fp = fopen(errfile, "r");

    if (!fp) {
        if (vflag)
            printf("No error file found, assuming no errors...\n");
    } else {
        fscanf(fp, "%d", &nrerrors);
        fclose(fp);
    }

    unlink(errfile);
    return nrerrors;
}

int
main(int argc, char *argv[])
{
    char *src;
    int srcext;
    char filebuf[PATH_MAX];
    char *libdir = NULL;
    int ret = 1;
    char *forkdir;
    char *pramdir;

    forkdir = getenv("FORKDIR");
    if (!forkdir) {
        fprintf(stderr, "fcc: error no $FORKDIR set\n");
        goto out;
    }

    pramdir = getenv("PRAMDIR");
    if (!pramdir) {
        fprintf(stderr, "fcc: error no $PRAMDIR set\n");
        goto out;
    }

    /* setup thelibdir directory */
    libdir = malloc(strlen(forkdir) + strlen("/lib/") + 1);
    if (!libdir)
        goto out;
    sprintf(libdir, "%s/lib/", forkdir);
    if (access(libdir, F_OK | R_OK)) {
        fprintf(stderr, "fcc: can't access libdir: %s\n", libdir);
        goto out;
    }

    /* init compaz */
    compaz.len = strlen(forkdir) + strlen("/lib/rcc") + 1;
    compaz.arg = malloc(compaz.len * sizeof(char));
    if (!compaz.arg)
        goto out;
    sprintf(compaz.arg, "%s/lib/rcc", forkdir);
    if (access(compaz.arg, R_OK | X_OK | F_OK)) {
        fprintf(stderr, "fcc: can't access compiler: %s\n", compaz.arg);
        goto out;
    }

    /* init prepraz */
    argz_append(argzp(&prepraz), "gcc", 4);
    argz_append(argzp(&prepraz), "-E", 3);

    /* init adaptaz */
    argz_append(argzp(&adaptaz), "gcc", 4);
    argz_append(argzp(&adaptaz), "-E", 3);
    snprintf(filebuf, sizeof(filebuf), "-I%s", libdir);
    argz_add(argzp(&adaptaz), filebuf);
    argz_append(argzp(&adaptaz), "-x", 3);
    argz_append(argzp(&adaptaz), "c", 2);
    argz_append(argzp(&adaptaz), "-o", 3);

    /* init linkaz */
    linkaz.len = strlen(pramdir) + strlen("/bin/plink") + 1;
    linkaz.arg = malloc(linkaz.len * sizeof(char));
    if (!linkaz.arg)
        goto out;
    sprintf(linkaz.arg, "%s/bin/plink", pramdir);
    if (access(linkaz.arg, R_OK | X_OK | F_OK)) {
        fprintf(stderr, "fcc: can't access the linker: %s\n", linkaz.arg);
        goto out;
    }

    /* init prassaz */
    assaz.len = strlen(pramdir) + strlen("/bin/prass") + 1;
    assaz.arg = malloc(assaz.len * sizeof(char));
    if (!assaz.arg)
        goto out;
    sprintf(assaz.arg, "%s/bin/prass", pramdir);
    if (access(assaz.arg, R_OK | X_OK | F_OK)) {
        fprintf(stderr, "fcc: can't access the assambler: %s\n", assaz.arg);
        goto out;
    }

    doargs(argc, argv);

    argz_append(argzp(&prepraz), "-o", 3);
    argz_append(argzp(&assaz), "-o", 3);

    if (vflag)
        printf(
            "Fork95 Compiler fcc version %s by C.W.Kessler\n"
            "FORKDIR: %s\nPRAMDIR: %s\n",
            VERSION, forkdir, pramdir);

    ret = 0;
    for (src = srcaz.arg; src; src = argz_next(argz(&srcaz), src)) {
        srcext = strlen(src) - 1;
        clearerrors(src);

        /*
        if (Eflag)
            sprintf(prepr, "%s %s -o %s.i %s.c",
                    thepreprocessor, prepropts, src, src);
        else
            sprintf(prepr, "%s %s -o %s.f %s.c",
                    thepreprocessor, prepropts, src, src);
        */
        src[srcext] = Eflag ? 'i' : 'f';
        argz_append(argzp(&prepraz), src, srcext + 2);

        src[srcext] = 'c';
        argz_append(argzp(&prepraz), src, srcext + 2);

        ret = argz_system(&prepraz, vflag);
        if (ret)
            goto out;

        /* delete files from argz for the next iteration */
        prepraz.len -= 2 * (srcext + 2);

        /* nur Praeprozessor */
        if (Eflag)
            goto out;

        /* compile:
           sprintf(compil, "%s %s %s.f %s.s; rm %s.f",
                   thecompiler, compopts, src, src, src);
        */

        src[srcext] = 'f';
        argz_append(argzp(&compaz), src, srcext + 2);
        src[srcext] = 's';
        argz_append(argzp(&compaz), src, srcext + 2);

        ret = argz_system(&compaz, vflag);
        if (ret)
            goto out;

        compaz.len -= 2 * (srcext + 2);

        src[srcext] = 'f';
        unlink(src);

        /*000120*/
        if ((vflag < 2) &&  (ret = geterrors(src, vflag))) {
            printf(ret > 1
                   ? "There was %d error. Files %.*s{s,o,f} removed.\n"
                   : "There were %d errors. Files %.*s{s,o,f} removed.\n",
                   ret, srcext, src);

            // asm file name
            src[srcext] = 's';
            unlink(src);

            // link file name
            src[srcext] = 'o';
            unlink(src);

            // src file name
            src[srcext] = 'i';
            unlink(src);

            // Exit: ret > 0 compiler errors
            goto out;
        }

        /*
        snprintf(adapt, 512, "%s -I%s -x c -o %s.i %s.s",
                 thepreprocessor, thelibdir, src, src);
        */
        src[srcext] = 'i';
        argz_append(argzp(&adaptaz), src, srcext + 2);
        src[srcext] = 's';
        argz_append(argzp(&adaptaz), src, srcext + 2);

        ret = argz_system(&adaptaz, vflag);
        if (ret)
            goto out;

        adaptaz.len -= 2 * (srcext + 2);

        if (!Sflag)
            unlink(src);

        /*
        snprintf(prass, 512, "%s %s -o %s.o %s.i; rm %s.i",
                 theassembler, assopts, src, src, src);
        */
        src[srcext] = 'o';
        argz_append(argzp(&assaz), src, srcext + 2);
        src[srcext] = 'i';
        argz_append(argzp(&assaz), src, srcext + 2);

        ret = argz_system(&assaz, vflag);
        if (ret)
            goto out;

        assaz.len -= 2 * (srcext + 2);

        unlink(src);
    }

    /* Nicht linken. */
    if (cflag)
        goto out;

    {
        /* link:
            snprintf(,
                    "%s %s %s "
                    "%s/%s "
                    "%s/stdlib.o "
                    "%s/graphic.o "
                    "%s/string.o "
                    "%s/ctype.o "
                    "%s/io.o "
                    "%s/print.o "
                    "%s/%s "
                    "%s/math.o",
                    thelinker, linkopts, lnks,
                    thelibdir, (Tflag) ? "forklib2.o" : "forklib.o",
                    thelibdir,
                    thelibdir,
                    thelibdir,
                    thelibdir,
                    thelibdir,
                    thelibdir,
                    thelibdir, (Tflag) ? "async2.o" : "async.o",
                    thelibdir);
        */
        const char *libobjects[] = {
            (Tflag) ? "forklib2.o" : "forklib.o",
            "stdlib.o", "graphic.o", "string.o",
            "ctype.o", "io.o", "print.o",
            (Tflag) ? "async2.o" : "async.o",
            "math.o", NULL
        };

        argz_append(argzp(&linkaz), argz(&tolinkaz));
        add_linkfiles(&linkaz, &srcaz);
        argz_add_with_prefix(&linkaz, libdir, libobjects);

        ret = argz_system(&linkaz, vflag);
        if (ret)
            goto out;
    }

out:
    if (libdir)
        free(libdir);
    argz_clear(&compaz);
    argz_clear(&linkaz);
    argz_clear(&prepraz);
    argz_clear(&assaz);
    argz_clear(&tolinkaz);
    argz_clear(&srcaz);
    argz_clear(&adaptaz);


    return ret;
}
