#include <fork.h>
#include <string.h>
#include "mpi.h"
#include "mpi_ops.h"

#define LOC_MAX 10
#define GLOB_MAX 100

#define MATRIX_HEIGHT 8

sh FILE *fp;

float *x;
float *local_x;
int x_size;
float *y;
float *local_y;
int y_size;
float *z;
float *local_z;
int z_size;

float **A;

int myrank,p;
int actual_number=0;
sh fair_lock lock;


void output(float *vector, int vec_size) {
  int i;
  //printf("ID:%d | ", myrank);
  for(i=0; i<vec_size; i++)
    printf(" %.3f ", vector[i]);
  prS("\n");
}

void mat_output(float **A, int x_dim, int y_dim) {
  int i;
  for(i=0; i<y_dim;i++) output(A[i], x_dim);
}

void readVector(float *vec, int *vec_dim) {
  int old_number = actual_number;
  fscanf(fp, "%d ", vec_dim);
  for(actual_number; actual_number < old_number + *vec_dim; actual_number++)
    fscanf(fp, " %f", &(vec[actual_number-old_number])); 
}

void main(void) {
  int a, b, i,j;
  MPI_Init(NULL,NULL);
  fair_lock_init(&lock);
  x = (float *) shmalloc(100 * sizeof(float));
  y = (float *) shmalloc(100 * sizeof(float));
  z = (float *) shmalloc(100 * sizeof(float));
  local_x = (float *) shmalloc(100 * sizeof(float));
  local_y = (float *) shmalloc(100 * sizeof(float));
  local_z = (float *) shmalloc(100 * sizeof(float));
  
  A = (float **) shmalloc(100 * sizeof(float *));
  for(i=0; i<MATRIX_HEIGHT; i++) A[i] = (float *) shmalloc(100*sizeof(float));
   
  
  MPI_Comm_size(MPI_COMM_WORLD, &p);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  
  if(myrank == 0) {
    prS("\n");
    if(!(fp = fopen("input.val", "r"))) {
      printf("Error reading input file!\n");
      exit(1);
    }
    for(i=0; i<MATRIX_HEIGHT;i++) {
      readVector(A[i], &x_size);
      //output(x, x_size);
    }
    readVector(x, &x_size);
    fclose(fp);    
    printf("\n\nA = \n");    
    mat_output(A, x_size,MATRIX_HEIGHT);
    printf("\n\nx = \n");
    output(x, x_size);
    printf("\n\n");
  }
  
  MPI_Bcast(&x_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Scatter(x, x_size/p, MPI_FLOAT,
	      x, x_size/p, MPI_FLOAT,
	      0, MPI_COMM_WORLD);
  
  for(i=0; i<MATRIX_HEIGHT; i++) {
    z[i] = 0.0;
    MPI_Scatter(A[i], x_size/p, MPI_FLOAT,
		A[i], x_size/p, MPI_FLOAT,
	      0, MPI_COMM_WORLD);
    for(j=0; j<x_size/p; j++) {
      z[i] += x[j] * A[i][j];
    }
  }
  MPI_Reduce(z, y, MATRIX_HEIGHT, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
  //fair_lockup(&lock);
  //fair_unlock(&lock);
  MPI_Barrier(MPI_COMM_WORLD);
  if(!myrank) for(i=0; i<MATRIX_HEIGHT; i++) printf("y[%d] = %f\n",i, y[i]);
  MPI_Finalize();
}
